# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kapitan',
 'kapitan.dependency_manager',
 'kapitan.inputs',
 'kapitan.inputs.templates.components.other_component',
 'kapitan.reclass',
 'kapitan.reclass.doc.source',
 'kapitan.reclass.reclass',
 'kapitan.reclass.reclass.adapters',
 'kapitan.reclass.reclass.datatypes',
 'kapitan.reclass.reclass.datatypes.tests',
 'kapitan.reclass.reclass.output',
 'kapitan.reclass.reclass.storage',
 'kapitan.reclass.reclass.storage.mixed',
 'kapitan.reclass.reclass.storage.tests',
 'kapitan.reclass.reclass.storage.yaml_fs',
 'kapitan.reclass.reclass.storage.yaml_git',
 'kapitan.reclass.reclass.tests',
 'kapitan.reclass.reclass.utils',
 'kapitan.reclass.reclass.utils.tests',
 'kapitan.reclass.reclass.values',
 'kapitan.reclass.reclass.values.tests',
 'kapitan.refs',
 'kapitan.refs.secrets',
 'kapitan.remoteinventory',
 'kapitan.validator',
 'reclass',
 'reclass.doc.source',
 'reclass.reclass',
 'reclass.reclass.adapters',
 'reclass.reclass.datatypes',
 'reclass.reclass.datatypes.tests',
 'reclass.reclass.output',
 'reclass.reclass.storage',
 'reclass.reclass.storage.mixed',
 'reclass.reclass.storage.tests',
 'reclass.reclass.storage.yaml_fs',
 'reclass.reclass.storage.yaml_git',
 'reclass.reclass.tests',
 'reclass.reclass.utils',
 'reclass.reclass.utils.tests',
 'reclass.reclass.values',
 'reclass.reclass.values.tests']

package_data = \
{'': ['*'],
 'kapitan': ['lib/*'],
 'kapitan.inputs': ['templates/components/my_component/*',
                    'templates/inventory/classes/*',
                    'templates/inventory/targets/*',
                    'templates/templates/docs/*',
                    'templates/templates/scripts/*'],
 'kapitan.reclass': ['doc/*',
                     'examples/ansible/*',
                     'examples/classes/*',
                     'examples/classes/debian/*',
                     'examples/classes/debian/release/*',
                     'examples/classes/debian/suite/*',
                     'examples/classes/mail/*',
                     'examples/nodes/munich/*',
                     'examples/nodes/zurich/*',
                     'examples/salt/*',
                     'releasenotes/*',
                     'releasenotes/notes/*',
                     'test/model/default/*',
                     'test/model/default/classes/*',
                     'test/model/default/classes/lab/env/*',
                     'test/model/default/nodes/*',
                     'test/model/extensions/*',
                     'test/model/extensions/classes/*',
                     'test/model/extensions/classes/lab/env/*',
                     'test/model/extensions/classes/relative/*',
                     'test/model/extensions/classes/relative/nested/*',
                     'test/model/extensions/classes/relative/nested/deep/*',
                     'test/model/extensions/classes/relative/nested/dive/*',
                     'test/model/extensions/nodes/*'],
 'kapitan.reclass.reclass.tests': ['data/01/classes/*',
                                   'data/01/nodes/*',
                                   'data/02/classes/*',
                                   'data/02/classes/one/*',
                                   'data/02/classes/two/*',
                                   'data/02/nodes/*',
                                   'data/03/classes/*',
                                   'data/03/nodes/alpha/*',
                                   'data/03/nodes/beta/*',
                                   'data/04/classes/*',
                                   'data/04/nodes/alpha/*',
                                   'data/05/classes/components/*',
                                   'data/05/classes/components/cluster-autoscaler/*',
                                   'data/05/classes/components/core-dns/*',
                                   'data/05/classes/versions/*',
                                   'data/05/nodes/*',
                                   'data/06/classes/applications/*',
                                   'data/06/classes/environments/*',
                                   'data/06/nodes/*'],
 'reclass': ['doc/*',
             'examples/ansible/*',
             'examples/classes/*',
             'examples/classes/debian/*',
             'examples/classes/debian/release/*',
             'examples/classes/debian/suite/*',
             'examples/classes/mail/*',
             'examples/nodes/munich/*',
             'examples/nodes/zurich/*',
             'examples/salt/*',
             'releasenotes/*',
             'releasenotes/notes/*',
             'test/model/default/*',
             'test/model/default/classes/*',
             'test/model/default/classes/lab/env/*',
             'test/model/default/nodes/*',
             'test/model/extensions/*',
             'test/model/extensions/classes/*',
             'test/model/extensions/classes/lab/env/*',
             'test/model/extensions/classes/relative/*',
             'test/model/extensions/classes/relative/nested/*',
             'test/model/extensions/classes/relative/nested/deep/*',
             'test/model/extensions/classes/relative/nested/dive/*',
             'test/model/extensions/nodes/*'],
 'reclass.reclass.tests': ['data/01/classes/*',
                           'data/01/nodes/*',
                           'data/02/classes/*',
                           'data/02/classes/one/*',
                           'data/02/classes/two/*',
                           'data/02/nodes/*',
                           'data/03/classes/*',
                           'data/03/nodes/alpha/*',
                           'data/03/nodes/beta/*',
                           'data/04/classes/*',
                           'data/04/nodes/alpha/*',
                           'data/05/classes/components/*',
                           'data/05/classes/components/cluster-autoscaler/*',
                           'data/05/classes/components/core-dns/*',
                           'data/05/classes/versions/*',
                           'data/05/nodes/*',
                           'data/06/classes/applications/*',
                           'data/06/classes/environments/*',
                           'data/06/nodes/*']}

install_requires = \
['addict>=2.4.0,<3.0.0',
 'azure-identity>=1.12.0,<2.0.0',
 'azure-keyvault-keys>=4.7.0,<5.0.0',
 'boto3>=1.18.17,<2.0.0',
 'certifi',
 'cryptography>=3.4.7,<4.0.0',
 'gitdb>=4.0.10,<5.0.0',
 'gitpython>=3.1.30,<4.0.0',
 'google-api-python-client>=2.15.0,<3.0.0',
 'hvac>=0.11.0,<0.12.0',
 'jinja2>=3.0.1,<4.0.0',
 'jsonnet>=0.18.0,<0.19.0',
 'jsonschema>=4.17.3,<5.0.0',
 'kadet>=0.2.2,<0.3.0',
 'packaging>=23.0,<24.0',
 'python-gnupg>=0.4.7,<0.5.0',
 'python-magic>=0.4.27,<0.5.0',
 'pyyaml>=6.0,<7.0',
 'requests>=2.28.2,<3.0.0',
 'six>=1.16.0,<2.0.0',
 'toml>=0.10.2,<0.11.0',
 'typing-extensions>=4.0.0,<5.0.0',
 'yamllint>=1.29.0,<2.0.0']

extras_require = \
{'gojsonnet': ['gojsonnet>=0.17.0,<0.18.0'], 'test': ['docker>=5.0.0,<6.0.0']}

entry_points = \
{'console_scripts': ['kapitan = kapitan.cli:main']}

setup_kwargs = {
    'name': 'kapitan',
    'version': '0.31.1rc3',
    'description': 'Generic templated configuration management for Kubernetes, Terraform and other things',
    'long_description': '# Kapitan: advanced configuration management tool\n\n![Unit Tests](https://github.com/kapicorp/kapitan/actions/workflows/test.yml/badge.svg)\n![Python Version](https://img.shields.io/pypi/pyversions/kapitan)\n![Downloads](https://img.shields.io/pypi/dm/kapitan)\n![Docker Pulls](https://img.shields.io/docker/pulls/kapicorp/kapitan)\n[![Docker](https://github.com/kapicorp/kapitan/workflows/Docker%20Build%20and%20Push/badge.svg)](https://github.com/kapicorp/kapitan/actions?query=workflow%3A%22Docker+Build+and+Push%22)\n[![Releases](https://img.shields.io/github/release/kapicorp/kapitan.svg)](https://github.com/kapicorp/kapitan/releases)\n[![Docker Image Size](https://img.shields.io/docker/image-size/kapicorp/kapitan/latest.svg)](https://hub.docker.com/r/kapicorp/kapitan)\n\n<img src="docs/images/kapitan_logo.png" width="25">\n\n\n**`Kapitan`** aims to be your *one-stop tool* to help you manage the ever growing complexity of your configurations.\n\nJoin the community [`#kapitan`](https://kubernetes.slack.com/archives/C981W2HD3)\n\n## [**Official site**](https://kapitan.dev) <https://kapitan.dev>\n\n\n## [**Quick Start**](https://kapitan.dev/kapitan_overview/#quickstart)\n\n## Install Kapitan\n\n### Docker (recommended)\n\n```shell\ndocker run -t --rm -v $(pwd):/src:delegated kapicorp/kapitan -h\n```\n\nOn Linux you can add `-u $(id -u)` to `docker run` to preserve file permissions.\n\n### Pip\n\nKapitan needs Python 3.7.\n\n#### Install Python 3.7\n\n* Linux: `sudo apt-get update && sudo apt-get install -y python3.7-dev python3-pip python3-yaml git`\n* Mac: `brew install python3 libyaml git libmagic`\n\n#### Install Kapitan\n\nUser (`$HOME/.local/lib/python3.7/bin` on Linux or `$HOME/Library/Python/3.7/bin` on macOS):\n\n```shell\npip3 install --user --upgrade kapitan\n```\n\nSystem-wide (not recommended):\n\n```shell\nsudo pip3 install --upgrade kapitan\n```\n\n## Build Kapitan\n\n### Docker\n\nTo build a docker image for the architecture of your machine, run `docker build . -t you-kapitan-image`, and to build for a specific platform, add `--platform linux/arm64`.\n\nTo build a multi-platform image (as the CI does), follow [the docker multi-platform documentation](https://docs.docker.com/build/building/multi-platform/).\n\n## Related projects\n\n* [Tesoro](https://github.com/kapicorp/tesoro) - Kubernetes Admission Controller for Kapitan Secrets\n* [Kapitan Reference](https://github.com/kapicorp/kapitan-reference) - our reference repository to get started with Kapitan\n',
    'author': 'Ricardo Amaro',
    'author_email': 'ramaro@kapicorp.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/kapicorp/kapitan',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python3

# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

import logging
import os
import shutil
from distutils.dir_util import copy_tree

from kapitan.inputs.base import InputType

logger = logging.getLogger(__name__)


class Remove(InputType):
    def __init__(self, compile_path, search_paths, ref_controller):
        super().__init__("remove", compile_path, search_paths, ref_controller)

    def compile_file(self, file_path, compile_path, ext_vars, **kwargs):
        """
        Write items in path as plain rendered files to compile_path.
        path can be either a file or directory.
        """

        try:
            logger.debug("Removing %s", file_path)
            if os.path.isfile(file_path):
                os.remove(file_path)
            else:
                shutil.rmtree(file_path)
        except OSError as e:
            logger.exception("Input dir not removed. Error: %s", e)

    def default_output_type(self):
        # no output_type options for remove
        return None



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass (http://github.com/madduck/reclass)
#
# Copyright © 2007–14 martin f. krafft <madduck@madduck.net>
# Released under the terms of the Artistic Licence 2.0
#
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

class OutputterBase(object):

    def __init__(self):
        pass

    def dump(self, data, pretty_print=False):
        raise NotImplementedError("dump() method not implemented.")


class OutputLoader(object):

    def __init__(self, outputter):
        self._name = 'reclass.output.' + outputter + '_outputter'
        try:
            self._module = __import__(self._name, globals(), locals(), self._name)
        except ImportError:
            raise NotImplementedError()

    def load(self, attr='Outputter'):
        klass = getattr(self._module, attr, None)
        if klass is None:
            raise AttributeError('Outputter class {0} does not export "{1}"'.format(self._name, klass))
        return klass



#!/usr/bin/env python3

# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

"kapitan targets"
import json
import logging
import multiprocessing
import os
import shutil
import sys
import tempfile
import time
from collections import defaultdict
from functools import partial

import jsonschema
import yaml
from reclass.errors import NotFoundError, ReclassException

from kapitan import cached, defaults
from kapitan.dependency_manager.base import fetch_dependencies
from kapitan.errors import CompileError, InventoryError, KapitanError
from kapitan.inputs.copy import Copy
from kapitan.inputs.external import External
from kapitan.inputs.helm import Helm
from kapitan.inputs.jinja2 import Jinja2
from kapitan.inputs.jsonnet import Jsonnet
from kapitan.inputs.kadet import Kadet
from kapitan.inputs.remove import Remove
from kapitan.remoteinventory.fetch import fetch_inventories, list_sources
from kapitan.resources import inventory_reclass
from kapitan.utils import dictionary_hash, directory_hash, hashable_lru_cache
from kapitan.validator.kubernetes_validator import KubernetesManifestValidator

logger = logging.getLogger(__name__)


def compile_targets(
    inventory_path, search_paths, output_path, parallel, targets, labels, ref_controller, **kwargs
):
    """
    Searches and loads target files, and runs compile_target() on a
    multiprocessing pool with parallel number of processes.
    kwargs are passed to compile_target()
    """
    # temp_path will hold compiled items
    temp_path = tempfile.mkdtemp(suffix=".kapitan")
    # enable previously compiled items to be reference in other compile inputs
    search_paths.append(temp_path)
    temp_compile_path = os.path.join(temp_path, "compiled")
    dep_cache_dir = temp_path

    updated_targets = targets
    try:
        updated_targets = search_targets(inventory_path, targets, labels)
    except CompileError as e:
        logger.error(e)
        sys.exit(1)

    # If --cache is set
    if kwargs.get("cache"):
        additional_cache_paths = kwargs.get("cache_paths")
        generate_inv_cache_hashes(inventory_path, targets, additional_cache_paths)
        # to cache fetched dependencies and inventories
        dep_cache_dir = os.path.join(output_path, ".dependency_cache")
        os.makedirs(dep_cache_dir, exist_ok=True)

        if not targets:
            updated_targets = changed_targets(inventory_path, output_path)
            logger.debug("Changed targets since last compilation: %s", updated_targets)
            if len(updated_targets) == 0:
                logger.info("No changes since last compilation.")
                return

    pool = multiprocessing.Pool(parallel)

    try:
        rendering_start = time.time()

        # check if --fetch or --force-fetch is enabled
        force_fetch = kwargs.get("force_fetch", False)
        fetch = kwargs.get("fetch", False) or force_fetch

        # deprecated --force flag
        if kwargs.get("force", False):
            logger.info(
                "DeprecationWarning: --force is deprecated. Use --force-fetch instead of --force --fetch"
            )
            force_fetch = True

        if fetch:
            # skip classes that are not yet available
            target_objs = load_target_inventory(inventory_path, updated_targets, ignore_class_notfound=True)
        else:
            # ignore_class_notfound = False by default
            target_objs = load_target_inventory(inventory_path, updated_targets)

        # append "compiled" to output_path so we can safely overwrite it
        compile_path = os.path.join(output_path, "compiled")

        if not target_objs:
            raise CompileError("Error: no targets found")

        # fetch inventory
        if fetch:
            # new_source checks for new sources in fetched inventory items
            new_sources = list(set(list_sources(target_objs)) - cached.inv_sources)
            while new_sources:
                fetch_inventories(
                    inventory_path,
                    target_objs,
                    dep_cache_dir,
                    force_fetch,
                    pool,
                )
                cached.reset_inv()
                target_objs = load_target_inventory(
                    inventory_path, updated_targets, ignore_class_notfound=True
                )
                cached.inv_sources.update(new_sources)
                new_sources = list(set(list_sources(target_objs)) - cached.inv_sources)
            # reset inventory cache and load target objs to check for missing classes
            cached.reset_inv()
            target_objs = load_target_inventory(inventory_path, updated_targets, ignore_class_notfound=False)
        # fetch dependencies
        if fetch:
            fetch_dependencies(output_path, target_objs, dep_cache_dir, force_fetch, pool)
        # fetch targets which have force_fetch: true
        elif not kwargs.get("force_fetch", False):
            fetch_objs = []
            # iterate through targets
            for target in target_objs:
                try:
                    # get value of "force_fetch" property
                    dependencies = target["dependencies"]
                    # dependencies is still a list
                    for entry in dependencies:
                        force_fetch = entry["force_fetch"]
                        if force_fetch:
                            fetch_objs.append(target)
                except KeyError:
                    # targets may have no "dependencies" or "force_fetch" key
                    continue
            # fetch dependencies from targets with force_fetch set to true
            if fetch_objs:
                fetch_dependencies(output_path, fetch_objs, dep_cache_dir, True, pool)

        logger.info("Rendered inventory (%.2fs)", time.time() - rendering_start)

        worker = partial(
            compile_target,
            search_paths=search_paths,
            compile_path=temp_compile_path,
            ref_controller=ref_controller,
            inventory_path=inventory_path,
            globals_cached=cached.as_dict(),
            **kwargs,
        )

        # compile_target() returns None on success
        # so p is only not None when raising an exception
        [p.get() for p in pool.imap_unordered(worker, target_objs) if p]

        os.makedirs(compile_path, exist_ok=True)

        # if '-t' is set on compile or only a few changed, only override selected targets
        if updated_targets:
            for target in updated_targets:
                compile_path_target = os.path.join(compile_path, target)
                temp_path_target = os.path.join(temp_compile_path, target)

                os.makedirs(compile_path_target, exist_ok=True)

                shutil.rmtree(compile_path_target)
                shutil.copytree(temp_path_target, compile_path_target)
                logger.debug("Copied %s into %s", temp_path_target, compile_path_target)
        # otherwise override all targets
        else:
            shutil.rmtree(compile_path)
            shutil.copytree(temp_compile_path, compile_path)
            logger.debug("Copied %s into %s", temp_compile_path, compile_path)

        # validate the compiled outputs
        if kwargs.get("validate", False):
            validate_map = create_validate_mapping(target_objs, compile_path)
            worker = partial(
                schema_validate_kubernetes_output,
                cache_dir=kwargs.get("schemas_path", "./schemas"),
            )
            [p.get() for p in pool.imap_unordered(worker, validate_map.items()) if p]

        # Save inventory and folders cache
        save_inv_cache(compile_path, targets)
        pool.close()

    except ReclassException as e:
        if isinstance(e, NotFoundError):
            logger.error("Inventory reclass error: inventory not found")
        else:
            logger.error("Inventory reclass error: %s", e.message)
        raise InventoryError(e.message)
    except Exception as e:
        # if compile worker fails, terminate immediately
        pool.terminate()
        logger.debug("Compile pool terminated")
        # only print traceback for errors we don't know about
        if not isinstance(e, KapitanError):
            logger.exception("\nUnknown (Non-Kapitan) error occurred:\n")

        logger.error("\n")
        if kwargs.get("verbose"):
            logger.exception(e)
        else:
            logger.error(e)
        sys.exit(1)
    finally:
        # always wait for other worker processes to terminate
        pool.join()
        shutil.rmtree(temp_path)
        logger.debug("Removed %s", temp_path)


def generate_inv_cache_hashes(inventory_path, targets, cache_paths):
    """
    generates the hashes for the inventory per target and jsonnet/jinja2 folders for caching purposes
    struct: {
        inventory:
            <target>:
                classes: <sha256>
                parameters: <sha256>
        folder:
            components: <sha256>
            docs: <sha256>
            lib: <sha256>
            scripts: <sha256>
            ...
    }
    """
    inv = inventory_reclass(inventory_path)
    cached.inv_cache = {}
    cached.inv_cache["inventory"] = {}
    cached.inv_cache["folder"] = {}

    if targets:
        for target in targets:
            try:
                cached.inv_cache["inventory"][target] = {}
                cached.inv_cache["inventory"][target]["classes"] = dictionary_hash(
                    inv["nodes"][target]["classes"]
                )
                cached.inv_cache["inventory"][target]["parameters"] = dictionary_hash(
                    inv["nodes"][target]["parameters"]
                )
            except KeyError:
                raise CompileError("target not found: {}".format(target))
    else:
        for target in inv["nodes"]:
            cached.inv_cache["inventory"][target] = {}
            cached.inv_cache["inventory"][target]["classes"] = dictionary_hash(
                inv["nodes"][target]["classes"]
            )
            cached.inv_cache["inventory"][target]["parameters"] = dictionary_hash(
                inv["nodes"][target]["parameters"]
            )

            compile_obj = inv["nodes"][target]["parameters"]["kapitan"]["compile"]
            for obj in compile_obj:
                for input_path in obj["input_paths"]:
                    base_folder = os.path.dirname(input_path).split("/")[0]
                    if base_folder == "":
                        base_folder = os.path.basename(input_path).split("/")[0]

                    if base_folder not in cached.inv_cache["folder"].keys():
                        if os.path.exists(base_folder) and os.path.isdir(base_folder):
                            cached.inv_cache["folder"][base_folder] = directory_hash(base_folder)

                # Cache additional folders set by --cache-paths
                for path in cache_paths:
                    if path not in cached.inv_cache["folder"].keys():
                        if os.path.exists(path) and os.path.isdir(path):
                            cached.inv_cache["folder"][path] = directory_hash(path)

        # Most commonly changed but not referenced in input_paths
        for common in ("lib", "vendor", "secrets"):
            if common not in cached.inv_cache["folder"].keys():
                if os.path.exists(common) and os.path.isdir(common):
                    cached.inv_cache["folder"][common] = directory_hash(common)


def changed_targets(inventory_path, output_path):
    """returns a list of targets that have changed since last compilation"""
    targets = []
    inv = inventory_reclass(inventory_path)

    saved_inv_cache = None
    saved_inv_cache_path = os.path.join(output_path, "compiled/.kapitan_cache")
    if os.path.exists(saved_inv_cache_path):
        with open(saved_inv_cache_path, "r") as f:
            try:
                saved_inv_cache = yaml.safe_load(f)
            except Exception:
                raise CompileError("Failed to load kapitan cache: %s", saved_inv_cache_path)

    targets_list = list(inv["nodes"])

    # If .kapitan_cache doesn't exist or failed to load, recompile all targets
    if not saved_inv_cache:
        return targets_list
    else:
        for key, hash in cached.inv_cache["folder"].items():
            try:
                if hash != saved_inv_cache["folder"][key]:
                    logger.debug("%s folder hash changed, recompiling all targets", key)
                    return targets_list
            except KeyError:
                # Errors usually occur when saved_inv_cache doesn't contain a new folder
                # Recompile anyway to be safe
                return targets_list

        for target in targets_list:
            try:
                if (
                    cached.inv_cache["inventory"][target]["classes"]
                    != saved_inv_cache["inventory"][target]["classes"]
                ):
                    logger.debug("classes hash changed in %s, recompiling", target)
                    targets.append(target)
                elif (
                    cached.inv_cache["inventory"][target]["parameters"]
                    != saved_inv_cache["inventory"][target]["parameters"]
                ):
                    logger.debug("parameters hash changed in %s, recompiling", target)
                    targets.append(target)
            except KeyError:
                # Errors usually occur when saved_inv_cache doesn't contain a new target
                # Recompile anyway to be safe
                targets.append(target)

    return targets


def save_inv_cache(compile_path, targets):
    """save the cache to .kapitan_cache for inventories per target and folders"""
    if cached.inv_cache:
        inv_cache_path = os.path.join(compile_path, ".kapitan_cache")
        # If only some targets were selected (-t), overwride only their inventory
        if targets:
            saved_inv_cache = None
            try:
                with open(inv_cache_path, "r") as f:
                    saved_inv_cache = yaml.safe_load(f)
            except Exception:
                pass

            if saved_inv_cache:
                if "inventory" not in saved_inv_cache:
                    saved_inv_cache["inventory"] = {}
            else:
                saved_inv_cache = {}
                saved_inv_cache["inventory"] = {}

            for target in targets:
                if target not in saved_inv_cache["inventory"]:
                    saved_inv_cache["inventory"][target] = {}

                saved_inv_cache["inventory"][target]["classes"] = cached.inv_cache["inventory"][target][
                    "classes"
                ]
                saved_inv_cache["inventory"][target]["parameters"] = cached.inv_cache["inventory"][target][
                    "parameters"
                ]

            with open(inv_cache_path, "w") as f:
                logger.debug("Saved .kapitan_cache for targets: %s", targets)
                yaml.dump(saved_inv_cache, stream=f, default_flow_style=False)

        else:
            with open(inv_cache_path, "w") as f:
                logger.debug("Saved .kapitan_cache")
                yaml.dump(cached.inv_cache, stream=f, default_flow_style=False)


def load_target_inventory(inventory_path, targets, ignore_class_notfound=False):
    """returns a list of target objects from the inventory"""
    target_objs = []
    inv = inventory_reclass(inventory_path, ignore_class_notfound)

    # if '-t' is set on compile, only loop through selected targets
    if targets:
        targets_list = targets
    else:
        targets_list = inv["nodes"]

    for target_name in targets_list:
        try:
            inv_target = inv["nodes"][target_name]
            target_obj = inv_target["parameters"]["kapitan"]
            # check if parameters.kapitan is empty
            if not target_obj:
                raise InventoryError(
                    "InventoryError: {}: parameters.kapitan has no assignment".format(target_name)
                )
            target_obj["target_full_path"] = inv_target["__reclass__"]["node"].replace("./", "")
            require_compile = not ignore_class_notfound
            valid_target_obj(target_obj, require_compile)
            validate_matching_target_name(target_name, target_obj, inventory_path)
            logger.debug("load_target_inventory: found valid kapitan target %s", target_name)
            target_objs.append(target_obj)
        except KeyError:
            logger.debug("load_target_inventory: target %s has no kapitan compile obj", target_name)
            pass

    return target_objs


def search_targets(inventory_path, targets, labels):
    """returns a list of targets where the labels match, otherwise just return the original targets"""
    if not labels:
        return targets

    try:
        labels_dict = dict(label.split("=") for label in labels)
    except ValueError:
        raise CompileError(
            "Compile error: Failed to parse labels, should be formatted like: kapitan compile -l env=prod app=example"
        )

    targets_found = []
    inv = inventory_reclass(inventory_path)

    for target_name in inv["nodes"]:
        matched_all_labels = False
        for label, value in labels_dict.items():
            try:
                if inv["nodes"][target_name]["parameters"]["kapitan"]["labels"][label] == value:
                    matched_all_labels = True
                    continue
            except KeyError:
                logger.debug("search_targets: label %s=%s didn't match target %s", label, value, target_name)

            matched_all_labels = False
            break

        if matched_all_labels:
            targets_found.append(target_name)

    if len(targets_found) == 0:
        raise CompileError("No targets found with labels: {}".format(labels))

    return targets_found


def compile_target(target_obj, search_paths, compile_path, ref_controller, globals_cached=None, **kwargs):
    """Compiles target_obj and writes to compile_path"""
    start = time.time()
    compile_objs = target_obj["compile"]
    ext_vars = target_obj["vars"]
    target_name = ext_vars["target"]

    if globals_cached:
        cached.from_dict(globals_cached)

    use_go_jsonnet = kwargs.get("use_go_jsonnet", False)
    if use_go_jsonnet:
        logger.debug("Using go-jsonnet over jsonnet")

    for comp_obj in compile_objs:
        input_type = comp_obj["input_type"]
        output_path = comp_obj["output_path"]
        input_params = comp_obj.setdefault("input_params", {})

        if input_type == "jinja2":
            input_compiler = Jinja2(compile_path, search_paths, ref_controller, comp_obj)
        elif input_type == "jsonnet":
            input_compiler = Jsonnet(compile_path, search_paths, ref_controller, use_go=use_go_jsonnet)
        elif input_type == "kadet":
            input_compiler = Kadet(compile_path, search_paths, ref_controller, input_params=input_params)
        elif input_type == "helm":
            input_compiler = Helm(compile_path, search_paths, ref_controller, comp_obj)
        elif input_type == "copy":
            ignore_missing = comp_obj.get("ignore_missing", False)
            input_compiler = Copy(compile_path, search_paths, ref_controller, ignore_missing)
        elif input_type == "remove":
            input_compiler = Remove(compile_path, search_paths, ref_controller)
        elif input_type == "external":
            input_compiler = External(compile_path, search_paths, ref_controller)
            if "args" in comp_obj:
                input_compiler.set_args(comp_obj["args"])
            if "env_vars" in comp_obj:
                input_compiler.set_env_vars(comp_obj["env_vars"])
        else:
            err_msg = 'Invalid input_type: "{}". Supported input_types: jsonnet, jinja2, kadet, helm, copy, remove, external'
            raise CompileError(err_msg.format(input_type))

        input_compiler.make_compile_dirs(target_name, output_path, **kwargs)
        input_compiler.compile_obj(comp_obj, ext_vars, **kwargs)

    logger.info("Compiled %s (%.2fs)", target_obj["target_full_path"], time.time() - start)


@hashable_lru_cache
def valid_target_obj(target_obj, require_compile=True):
    """
    Validates a target_obj
    Returns a dict object if target is valid
    Otherwise raises ValidationError
    """

    schema = {
        "type": "object",
        "properties": {
            "vars": {"type": "object"},
            "secrets": {
                "type": "object",
                "properties": {
                    "gpg": {
                        "type": "object",
                        "properties": {
                            "recipients": {
                                "type": "array",
                                "items": {
                                    "type": "object",
                                    "properties": {
                                        "name": {"type": "string"},
                                        "fingerprint": {"type": "string"},
                                    },
                                },
                            },
                        },
                        "required": ["recipients"],
                    },
                    "gkms": {
                        "type": "object",
                        "properties": {"key": {"type": "string"}},
                        "required": ["key"],
                    },
                    "awskms": {
                        "type": "object",
                        "properties": {"key": {"type": "string"}},
                        "required": ["key"],
                    },
                    "azkms": {
                        "type": "object",
                        "properties": {"key": {"type": "string"}},
                        "required": ["key"],
                    },
                    "vaultkv": {
                        "type": "object",
                        "properties": {
                            "VAULT_ADDR": {"type": "string"},
                            "VAULT_NAMESPACE": {"type": "string"},
                            "VAULT_SKIP_VERIFY": {"type": "string"},
                            "VAULT_CLIENT_KEY": {"type": "string"},
                            "VAULT_CLIENT_CERT": {"type": "string"},
                            "auth": {"enum": ["token", "userpass", "ldap", "github", "approle"]},
                            "engine": {"type": "string"},
                            "mount": {"type": "string"},
                        },
                    },
                    "vaulttransit": {
                        "type": "object",
                        "properties": {
                            "VAULT_ADDR": {"type": "string"},
                            "VAULT_NAMESPACE": {"type": "string"},
                            "VAULT_SKIP_VERIFY": {"type": "string"},
                            "VAULT_CLIENT_KEY": {"type": "string"},
                            "VAULT_CLIENT_CERT": {"type": "string"},
                            "auth": {"enum": ["token", "userpass", "ldap", "github", "approle"]},
                            "engine": {"type": "string"},
                            "mount": {"type": "string"},
                            "key": {"type": "string"},
                        },
                    },
                },
                "additionalProperties": False,
            },
            "compile": {
                "type": "array",
                "items": {
                    "type": "object",
                    "properties": {
                        "name": {"type": "string"},
                        "input_paths": {"type": "array"},
                        "input_type": {"type": "string"},
                        "output_path": {"type": "string"},
                        "output_type": {"type": "string"},
                        "helm_values": {"type": "object"},
                        "helm_values_files": {"type": "array"},
                        "helm_params": {
                            "type": "object",
                            "properties": {"name": {"type": "string"}},
                            "additionalProperties": True,
                        },
                        "input_params": {"type": "object"},
                        "env_vars": {"type": "object"},
                        "args": {"type": "array"},
                        "suffix_remove": {"type": "boolean"},
                        "suffix_stripped": {"type": "string"},
                    },
                    "required": ["input_type", "input_paths", "output_path"],
                    "minItems": 1,
                    "oneOf": [
                        {
                            "properties": {
                                "input_type": {"enum": ["jsonnet", "kadet", "copy", "remove"]},
                                "output_type": {"enum": ["yml", "yaml", "json", "plain", "toml"]},
                            },
                        },
                        {"properties": {"input_type": {"enum": ["jinja2", "helm", "external"]}}},
                    ],
                },
            },
            "validate": {
                "type": "array",
                "items": {
                    "type": "object",
                    "properties": {
                        "output_paths": {"type": "array"},
                        "type": {"type": "string", "enum": ["kubernetes"]},
                        "kind": {"type": "string"},
                        "version": {"type": "string"},
                    },
                    "required": ["output_paths", "type"],
                    "minItems": 1,
                    "allOf": [
                        {
                            "if": {"properties": {"type": {"const": "kubernetes"}}},
                            "then": {
                                "properties": {
                                    "type": {},
                                    "kind": {},
                                    "output_paths": {},
                                    "version": {},
                                },
                                "additionalProperties": False,
                                "required": ["kind"],
                            },
                        },
                    ],
                },
            },
            "dependencies": {
                "type": "array",
                "items": {
                    "type": "object",
                    "properties": {
                        "chart_name": {"type": "string"},
                        "type": {"type": "string", "enum": ["git", "http", "https", "helm"]},
                        "output_path": {"type": "string"},
                        "source": {"type": "string"},
                        "subdir": {"type": "string"},
                        "ref": {"type": "string"},
                        "unpack": {"type": "boolean"},
                        "version": {"type": "string"},
                        "force_fetch": {"type": "boolean"},
                    },
                    "required": ["type", "output_path", "source"],
                    "additionalProperties": False,
                    "allOf": [
                        {
                            "if": {"properties": {"type": {"enum": ["http", "https"]}}},
                            "then": {
                                "properties": {
                                    "type": {},
                                    "source": {"format": "uri"},
                                    "output_path": {},
                                    "unpack": {},
                                    "force_fetch": {},
                                },
                                "additionalProperties": False,
                            },
                        },
                        {
                            "if": {"properties": {"type": {"enum": ["helm"]}}},
                            "then": {
                                "properties": {
                                    "type": {},
                                    "source": {"format": "uri"},
                                    "output_path": {},
                                    "unpack": {},
                                    "chart_name": {"type": "string"},
                                    "version": {"type": "string"},
                                    "force_fetch": {},
                                },
                                "required": ["type", "output_path", "source", "chart_name"],
                                "additionalProperties": False,
                            },
                        },
                    ],
                },
            },
            "inventory": {
                "type": "array",
                "items": {
                    "type": "object",
                    "properties": {
                        "type": {"type": "string", "enum": ["git", "http", "https"]},
                        "output_path": {"type": "string"},
                        "source": {"type": "string"},
                        "subdir": {"type": "string"},
                        "ref": {"type": "string"},
                        "unpack": {"type": "boolean"},
                    },
                    "required": ["type", "output_path", "source"],
                    "additionalProperties": False,
                    "allOf": [
                        {
                            "if": {"properties": {"type": {"enum": ["http", "https"]}}},
                            "then": {
                                "properties": {
                                    "type": {},
                                    "source": {"format": "uri"},
                                    "output_path": {},
                                    "unpack": {},
                                },
                                "additionalProperties": False,
                            },
                        },
                    ],
                },
            },
        },
    }
    if require_compile:
        schema["required"] = ["compile"]

    try:
        jsonschema.validate(target_obj, schema, format_checker=jsonschema.FormatChecker())
    except jsonschema.exceptions.ValidationError as e:
        raise InventoryError(
            "Invalid inventory structure\n\nError: {}\nOn instance:\n{}".format(
                e.message, json.dumps(e.instance, indent=2, sort_keys=False)
            )
        )


def validate_matching_target_name(target_filename, target_obj, inventory_path):
    """Throws *InventoryError* if parameters.kapitan.vars.target is not set,
    or target does not have a corresponding yaml file in *inventory_path*
    """
    logger.debug("validating target name matches the name of yml file %s", target_filename)
    try:
        target_name = target_obj["vars"]["target"]
    except KeyError:
        error_message = (
            f'Target missing: target "{target_filename}" is missing parameters.kapitan.vars.target\n'
            "This parameter should be set to the target name"
        )
        raise InventoryError(error_message)

    if target_filename != target_name:
        target_path = os.path.join(os.path.abspath(inventory_path), "targets")

        error_message = (
            f'Target "{target_name}" is missing the corresponding yml file in {target_path}\n'
            "Target name should match the name of the target yml file in inventory"
        )
        raise InventoryError(error_message)


def schema_validate_compiled(args):
    """
    validates compiled output according to schemas specified in the inventory
    """
    if not os.path.isdir(args.compiled_path):
        logger.error("compiled-path %s not found", args.compiled_path)
        sys.exit(1)

    if not os.path.isdir(args.schemas_path):
        os.makedirs(args.schemas_path)
        logger.info("created schema-cache-path at %s", args.schemas_path)

    worker = partial(schema_validate_kubernetes_output, cache_dir=args.schemas_path)
    pool = multiprocessing.Pool(args.parallelism)

    try:
        target_objs = load_target_inventory(args.inventory_path, args.targets)
        validate_map = create_validate_mapping(target_objs, args.compiled_path)

        [p.get() for p in pool.imap_unordered(worker, validate_map.items()) if p]
        pool.close()

    except ReclassException as e:
        if isinstance(e, NotFoundError):
            logger.error("Inventory reclass error: inventory not found")
        else:
            logger.error("Inventory reclass error: %s", e.message)
        raise InventoryError(e.message)
    except Exception as e:
        pool.terminate()
        logger.debug("Validate pool terminated")
        # only print traceback for errors we don't know about
        if not isinstance(e, KapitanError):
            logger.exception("Unknown (Non-Kapitan) Error occured")

        logger.error("\n")
        logger.error(e)
        sys.exit(1)
    finally:
        # always wait for other worker processes to terminate
        pool.join()


def create_validate_mapping(target_objs, compiled_path):
    """
    creates mapping of (kind, version) tuple to output_paths across different targets
    this is required to avoid redundant schema fetch when multiple targets use the same schema for validation
    """
    validate_files_map = defaultdict(list)
    for target_obj in target_objs:
        target_name = target_obj["vars"]["target"]
        if "validate" not in target_obj:
            logger.debug(
                "target '%s' does not have 'validate' parameter in inventory. skipping",
                target_name,
            )
            continue

        for validate_item in target_obj["validate"]:
            validate_type = validate_item["type"]
            if validate_type == "kubernetes":
                kind_version_pair = (
                    validate_item["kind"],
                    validate_item.get("version", defaults.DEFAULT_KUBERNETES_VERSION),
                )
                for output_path in validate_item["output_paths"]:
                    full_output_path = os.path.join(compiled_path, target_name, output_path)
                    if not os.path.isfile(full_output_path):
                        logger.warning(
                            "%s does not exist for target '%s'. skipping", output_path, target_name
                        )
                        continue
                    validate_files_map[kind_version_pair].append(full_output_path)
            else:
                logger.warning("type %s is not supported for validation. skipping", validate_type)

    return validate_files_map


def schema_validate_kubernetes_output(validate_data, cache_dir):
    """
    validates given files according to kubernetes manifest schemas
    schemas are cached from/to cache_dir
    validate_data must be of structure ((kind, version), validate_files)
    """
    (kind, version), validate_files = validate_data
    KubernetesManifestValidator(cache_dir).validate(validate_files, kind=kind, version=version)



class ParameterDict(dict):
    def __init__(self, *args, **kwargs):
        self._uri = kwargs.pop('uri', None)
        dict.__init__(self, *args, **kwargs)

    @property
    def uri(self):
        return self._uri

    @uri.setter
    def uri(self, uri):
        self._uri = uri



#!/usr/bin/env python3

# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

import logging
import os
import tempfile

import yaml

from kapitan.errors import CompileError, HelmTemplateError
from kapitan.helm_cli import helm_cli
from kapitan.inputs.base import CompiledFile, InputType
from kapitan.inputs.kadet import BaseModel, BaseObj, Dict

logger = logging.getLogger(__name__)

HELM_DENIED_FLAGS = {
    "dry-run",
    "generate-name",
    "help",
    "output-dir",
    "show-only",
}

HELM_DEFAULT_FLAGS = {"--include-crds": True, "--skip-tests": True}


class Helm(InputType):
    def __init__(self, compile_path, search_paths, ref_controller, args):
        super().__init__("helm", compile_path, search_paths, ref_controller)

        self.helm_values_files = args.get("helm_values_files")
        self.helm_params = args.get("helm_params") or {}
        self.helm_path = args.get("helm_path")
        self.file_path = None

        self.helm_values_file = None
        if "helm_values" in args:
            self.helm_values_file = write_helm_values_file(args["helm_values"])

        self.kube_version = None
        if "kube_version" in args:
            logger.warning("passing kube_version is deprecated. Use api_versions helm flag instead.")
            self.kube_version = args["kube_version"]

    def compile_file(self, file_path, compile_path, ext_vars, **kwargs):
        """
        Render templates in file_path/templates and write to compile_path.
        file_path must be a directory containing helm chart.
        kwargs:
            reveal: default False, set to reveal refs on compile
            target_name: default None, set to current target being compiled
        """
        reveal = kwargs.get("reveal", False)
        target_name = kwargs.get("target_name", None)
        indent = kwargs.get("indent", 2)

        if self.file_path is not None:
            raise CompileError(
                "The same helm input was compiled with different input paths, which will give a wrong result."
                + f" The input paths found are: {self.file_path} and {file_path}."
                + f" The search paths were: {self.search_paths}."
            )
        self.file_path = file_path

        temp_dir = tempfile.mkdtemp()
        os.makedirs(os.path.dirname(compile_path), exist_ok=True)
        # save the template output to temp dir first
        _, error_message = self.render_chart(
            chart_dir=file_path,
            output_path=temp_dir,
            helm_path=self.helm_path,
            helm_params=self.helm_params,
            helm_values_file=self.helm_values_file,
            helm_values_files=self.helm_values_files,
        )
        if error_message:
            raise HelmTemplateError(error_message)

        walk_root_files = os.walk(temp_dir)
        for current_dir, _, files in walk_root_files:
            for file in files:  # go through all the template files
                rel_dir = os.path.relpath(current_dir, temp_dir)
                rel_file_name = os.path.join(rel_dir, file)
                full_file_name = os.path.join(current_dir, file)
                with open(full_file_name, "r") as f:
                    item_path = os.path.join(compile_path, rel_file_name)
                    os.makedirs(os.path.dirname(item_path), exist_ok=True)
                    with CompiledFile(
                        item_path,
                        self.ref_controller,
                        mode="w",
                        reveal=reveal,
                        target_name=target_name,
                        indent=indent,
                    ) as fp:
                        yml_obj = list(yaml.safe_load_all(f))
                        fp.write_yaml(yml_obj)
                        logger.debug("Wrote file %s to %s", full_file_name, item_path)

        self.helm_values_file = None  # reset this
        self.helm_params = {}
        self.helm_values_files = []

    def default_output_type(self):
        return None

    def render_chart(
        self, chart_dir, output_path, helm_path, helm_params, helm_values_file, helm_values_files
    ):
        helm_flags = dict(HELM_DEFAULT_FLAGS)
        # add to flags if set
        if self.kube_version:
            helm_flags["--api-versions"] = self.kube_version

        return render_chart(
            chart_dir,
            output_path,
            helm_path,
            helm_params,
            helm_values_file,
            helm_values_files,
            helm_flags=helm_flags,
        )


def render_chart(
    chart_dir,
    output_path,
    helm_path,
    helm_params,
    helm_values_file,
    helm_values_files,
    helm_flags=None,
):
    """
    Renders chart in chart_dir
    Returns tuple with output and error_message
    """
    args = ["template"]

    name = helm_params.pop("name", None)
    output_file = helm_params.pop("output_file", None)

    if helm_flags is None:
        helm_flags = HELM_DEFAULT_FLAGS

    for param, value in helm_params.items():
        if len(param) == 1:
            raise ValueError(f"invalid helm flag: '{param}'. helm_params supports only long flag names")

        if "-" in param:
            raise ValueError(f"helm flag names must use '_' and not '-': {param}")

        param = param.replace("_", "-")

        if param in ("set", "set-file", "set-string"):
            raise ValueError(
                f"helm '{param}' flag is not supported. Use 'helm_values' to specify template values"
            )

        if param == "values":
            raise ValueError(
                f"helm '{param}' flag is not supported. Use 'helm_values_files' to specify template values files"
            )

        if param in HELM_DENIED_FLAGS:
            raise ValueError(f"helm flag '{param}' is not supported.")

        helm_flags[f"--{param}"] = value

    # 'release_name' used to be the "helm template" [NAME] parameter.
    # For backward compatibility, assume it is the '--release-name' flag only if its value is a bool.
    release_name = helm_flags.get("--release-name")
    if release_name is not None and not isinstance(release_name, bool):
        logger.warning("using 'release_name' to specify the output name is deprecated. Use 'name' instead")
        del helm_flags["--release-name"]
        # name is used in place of release_name if both are specified
        name = name or release_name

    for flag, value in helm_flags.items():
        # boolean flag should be passed when present, and omitted when not specified
        if isinstance(value, bool):
            if value:
                args.append(flag)
        else:
            args.append(flag)
            args.append(str(value))

    """renders helm chart located at chart_dir, and stores the output to output_path"""
    if helm_values_file:
        args.append("--values")
        args.append(helm_values_file)

    if helm_values_files:
        for file_name in helm_values_files:
            args.append("--values")
            args.append(file_name)

    if not output_file and output_path not in (None, "-"):
        args.append("--output-dir")
        args.append(output_path)

    if "name_template" not in helm_flags:
        args.append(name or "--generate-name")

    # uses absolute path to make sure helm interprets it as a
    # local dir and not a chart_name that it should download.
    args.append(chart_dir)

    # If output_path is '-', output is a string with rendered chart
    if output_path == "-":
        _, helm_output = tempfile.mkstemp(".helm_output.yml", text=True)
        with open(helm_output, "w+") as f:
            error_message = helm_cli(helm_path, args, stdout=f)
            f.seek(0)
            return (f.read(), error_message)

    if output_file:
        with open(os.path.join(output_path, output_file), "wb") as f:
            # can't be verbose when capturing stdout
            error_message = helm_cli(helm_path, args, stdout=f)
            return ("", error_message)
    else:
        error_message = helm_cli(helm_path, args, verbose="--debug" in helm_flags)
        return ("", error_message)


def write_helm_values_file(helm_values: dict):
    """
    Dump helm values into a yaml file whose path will
    be passed over to helm binary
    """
    _, helm_values_file = tempfile.mkstemp(".helm_values.yml", text=True)
    with open(helm_values_file, "w") as fp:
        yaml.safe_dump(helm_values, fp)

    return helm_values_file


class HelmChart(BaseModel):
    """
    Returns rendered helm chart in chart_dir
    Each rendered file will be a key in self.root

    Requires chart_dir to exist (it will not download it)
    """

    chart_dir: str
    helm_params: dict = {}
    helm_values: dict = {}
    helm_path: str = None

    def new(self):
        for obj in self.load_chart():
            self.root[
                f"{obj['metadata']['name'].lower()}-{obj['kind'].lower().replace(':','-')}"
            ] = BaseObj.from_dict(obj)

    def load_chart(self):
        helm_values_file = None
        if self.helm_values != {}:
            helm_values_file = write_helm_values_file(self.helm_values)
        output, error_message = render_chart(
            self.chart_dir, "-", self.helm_path, self.helm_params, helm_values_file, None
        )
        if error_message:
            raise HelmTemplateError(error_message)

        return yaml.safe_load_all(output)
